"""Utilities to convert a file tree."""

from __future__ import annotations

import collections
import dataclasses
import multiprocessing.pool
import pathlib
import typing

from . import path


class Rule(typing.Protocol):
    def match(self, source_relative_path: pathlib.Path) -> bool:
        ...

    def skip(self) -> bool:
        ...

    def rename(self, name: str) -> str:
        ...


class Rename:
    def __init__(
        self,
        match_relative_path: typing.Union[str, pathlib.Path],
        new_name: str,
    ):
        self.match_relative_path = pathlib.Path(match_relative_path)
        self.new_name = new_name

    def match(self, source_relative_path: pathlib.Path) -> bool:
        return source_relative_path == self.match_relative_path

    def skip(self) -> bool:
        return False

    def rename(self, name: str) -> str:
        return self.new_name


class RenameExtension:
    def __init__(self, match_extension: str, new_extension: str):
        self.match_extension = match_extension
        self.new_extension = new_extension

    def match(self, source_relative_path: pathlib.Path) -> bool:
        return source_relative_path.suffix == self.match_extension

    def skip(self) -> bool:
        return False

    def rename(self, name: str) -> str:
        name_as_path = pathlib.Path(name)
        return f"{name_as_path.stem}{self.new_extension}"


class SkipName:
    def __init__(
        self,
        match_name: str,
    ):
        self.match_name = match_name

    def match(self, source_relative_path: pathlib.Path) -> bool:
        return source_relative_path.name == self.match_name

    def skip(self) -> bool:
        return True

    def rename(self, name: str) -> str:
        return name


@dataclasses.dataclass
class Task:
    source: pathlib.Path
    target_directory: path.Directory
    target_name: str
    index: int
    total: int


def tasks(
    source: pathlib.Path,
    target: path.Directory,
    rules: list[Rule],
    source_root: pathlib.Path,
    target_root: pathlib.Path,
    debouncer: typing.Optional[path.Debouncer],
) -> typing.Iterator[Task]:
    node_sources = list(source.iterdir())
    node_sources.sort(
        key=lambda node_source: f"1{node_source.name}"
        if node_source.is_dir()
        else f"2{node_source.name}"
    )
    directories_nodes: list[tuple[pathlib.Path, str]] = []
    for node_source in node_sources:
        source_relative_path = node_source.relative_to(source_root)
        target_name = node_source.name
        skip = False
        for rule in rules:
            if rule.match(source_relative_path):
                if rule.skip():
                    skip = True
                    break
                target_name = rule.rename(target_name)
        if skip:
            continue
        if node_source.is_dir():
            directories_nodes.append((node_source, target_name))
        else:
            yield Task(
                source=node_source,
                target_directory=target,
                target_name=target_name,
                index=0,
                total=0,
            )
    for node_source, target_name in directories_nodes:
        yield from tasks(
            source=node_source,
            target=target.create_subdirectory(target_name, debouncer=debouncer),
            rules=rules,
            source_root=source_root,
            target_root=target_root,
            debouncer=debouncer,
        )


def copy_tree(
    source: pathlib.Path,
    target: pathlib.Path,
    rules: list[Rule],
    handle_task: typing.Callable[[Task], None],
    thread_pool_processes: typing.Optional[int] = None,
    pool_imap_chunksize: int = 1,
):
    with path.Debouncer() as debouncer:
        target_directory = path.Directory(path=target, debouncer=debouncer)
        tasks_list = list(
            tasks(
                source=source,
                target=target_directory,
                rules=rules,
                source_root=source,
                target_root=target,
                debouncer=debouncer,
            )
        )
        for index, task in enumerate(tasks_list):
            task.index = index
            task.total = len(tasks_list)
        with multiprocessing.pool.ThreadPool(processes=thread_pool_processes) as pool:
            collections.deque(
                pool.map(
                    handle_task,
                    tasks_list,
                    chunksize=pool_imap_chunksize,
                ),
                maxlen=0,
            )



"""File system wrappers."""

from __future__ import annotations

import bisect
import copy
import itertools
import json
import operator
import pathlib
import threading
import time
import types
import typing

import brotli
import numpy
import undr

DEFAULT_INDEX_DATA = {
    "directories": [],
    "files": [],
    "other_files": [],
    "version": {"major": 1, "minor": 0, "patch": 0},
}


class Type(typing.Protocol):
    """A file type supported by UNDR."""

    def dtype(self) -> numpy.dtype:
        ...

    def extension(self) -> str:
        ...

    def properties(self) -> dict[str, typing.Any]:
        ...


class DvsType:
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    def dtype(self) -> numpy.dtype:
        return undr.raw.DVS_DTYPE

    def extension(self) -> str:
        return "dvs"

    def properties(self) -> dict[str, typing.Any]:
        return {
            "type": "dvs",
            "width": self.width,
            "height": self.height,
        }


class ApsType:
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    def dtype(self) -> numpy.dtype:
        return undr.raw.aps_dtype(width=self.width, height=self.height)

    def extension(self) -> str:
        return "aps"

    def properties(self) -> dict[str, typing.Any]:
        return {
            "type": "aps",
            "width": self.width,
            "height": self.height,
        }


class ImuType:
    def dtype(self) -> numpy.dtype:
        return undr.raw.IMU_DTYPE

    def extension(self) -> str:
        return "imu"

    def properties(self) -> dict[str, typing.Any]:
        return {
            "type": "imu",
        }


class Directory:
    """A local directory.

    Args:
        path (typing.Union[pathlib.Path, str]): The path of the directory in the local file system.
        debouncer (typing.Optional[Debouncer], optional): Delegate that saves the index to the disk to optimize frequent writes. Defaults to None.
    """

    def __init__(
        self,
        path: typing.Union[pathlib.Path, str],
        debouncer: typing.Optional["Debouncer"] = None,
    ):
        if isinstance(path, str):
            self.path = pathlib.Path(path)
        else:
            self.path = path
        self.path.mkdir(exist_ok=True, parents=False)
        self.index_path = self.path / "-index.json"
        if not self.index_path.exists():
            default_index_content = (
                f"{json.dumps(DEFAULT_INDEX_DATA, sort_keys=True, indent=4)}\n".encode()
            )
            with open(self.index_path, "wb") as index_data_file:
                index_data_file.write(default_index_content)
        self.index_data = undr.json_index.load(self.index_path)
        names = set()
        for directory in self.index_data["directories"]:
            if directory in names:
                raise Exception(
                    f'duplicate file or directory "{directory}" in {self.index_path}'
                )
            names.add(directory)
        self.index_data["directories"].sort()
        for file in itertools.chain(
            self.index_data["files"], self.index_data["other_files"]
        ):
            if file["name"] in names:
                raise Exception(
                    f"duplicate file or directory \"{file['name']}\" in {self.index_path}"
                )
            names.add(file["name"])
        self.index_data["files"].sort(key=operator.itemgetter("name"))
        self.index_data["other_files"].sort(key=operator.itemgetter("name"))
        self.added_names = set()
        self.index_update_lock = threading.Lock()
        self.index_write_lock = threading.Lock()
        self.debouncer = debouncer
        if self.debouncer is not None:
            self.debouncer_id = self.debouncer.register(self)
        else:
            self.debouncer_id = 0

    def create_subdirectory(
        self, name: str, debouncer: typing.Optional["Debouncer"] = None
    ) -> "Directory":
        with self.index_update_lock:
            if name in self.added_names:
                raise Exception(
                    f'attempted to add the file or directory "{name}" twice in {self.index_path}'
                )
            self.added_names.add(name)
            insertion_point = bisect.bisect_left(self.index_data["directories"], name)
            if insertion_point == len(self.index_data["directories"]):
                self.index_data["directories"].append(name)
            elif name != self.index_data["directories"][insertion_point]:
                self.index_data["directories"].insert(insertion_point, name)
            if self.debouncer is None:
                self.save_index_data(index_update_lock_already_acquired=True)
            else:
                self.debouncer.set(self.debouncer_id)
        return Directory(path=self.path / name, debouncer=debouncer)

    def create_file(
        self, type: Type, name: str, metadata: dict[str, typing.Any]
    ) -> "File":
        """Adds a file to this directory.

        The file is a context manager compatible with the 'with' statement.
        Upon exiting the context manaager, the file calls :py:meth:`update_index`
        to register itself in the index of this directory.
        If a file with the same name exists, it is silently replaced.

        This function is thread-safe.

        Args:
            type (Type): A type object compatible with UNDR (:py:class:`DvsType`, :py:class:`ApsType`, or :py:class:`ImuType`).
            name (str): The file name without extension (the extension is automatically set from the type).
            metadata (dict[str, typing.Any]): File properties to add to the index.

        Returns:
            File: A writable file.
        """
        return File(directory=self, type=type, name=name, metadata=metadata)

    def create_other_file(
        self, name: str, metadata: dict[str, typing.Any]
    ) -> "OtherFile":
        """Adds a file to this directory.

        The file is a context manager compatible with the 'with' statement.
        Upon exiting the context manaager, the file calls :py:meth:`update_index`
        to register itself in the index of this directory.
        If a file with the same name exists, it is silently replaced.

        This function is thread-safe.

        Args:
            name (str): The file name including the extension.

        Returns:
            OtherFile: A writable file.
        """
        return OtherFile(directory=self, name=name, metadata=metadata)

    def update_index(self, file: "BaseFile"):
        """Adds a file to the index.

        This function is called automatically by files used as context managers ('with').
        It is thread-safe.

        Args:
            file (str): The file to add to the index.
        """
        with self.index_update_lock:
            if isinstance(file, File):
                key = "files"
            elif isinstance(file, OtherFile):
                key = "other_files"
            else:
                raise Exception(f"unsupported base type for file {file._path}")
            if file._index_entry["name"] in self.added_names:
                raise Exception(
                    f"attempted to add the file or directory \"{file._index_entry['name']}\" twice in {self.index_path}"
                )
            self.added_names.add(file._index_entry["name"])
            assert file._index_entry["hash"] is not None
            insertion_point = bisect.bisect_left(
                self.index_data[key],
                file._index_entry["name"],
                key=operator.itemgetter("name"),
            )
            if insertion_point == len(self.index_data[key]):
                self.index_data[key].append(file._index_entry)
            elif (
                self.index_data[key][insertion_point]["name"]
                == file._index_entry["name"]
            ):
                self.index_data[key][insertion_point] = file._index_entry
            else:
                self.index_data[key].insert(insertion_point, file._index_entry)
            if self.debouncer is None:
                self.save_index_data(index_update_lock_already_acquired=True)
            else:
                self.debouncer.set(self.debouncer_id)

    def save_index_data(self, index_update_lock_already_acquired: bool):
        """Write the index to the disk.

        This function is called automatically by files used as context managers ('with').
        It is thread-safe.

        Args:
            index_update_lock_already_acquired bool): Whether the caller has already acquired ``index_update_lock``.
        """
        if index_update_lock_already_acquired:
            index_data = copy.deepcopy(self.index_data)
            self.index_write_lock.acquire()
        else:
            with self.index_update_lock:
                index_data = copy.deepcopy(self.index_data)
                self.index_write_lock.acquire()
        try:
            index_content = (
                f"{json.dumps(index_data, sort_keys=True, indent=4)}\n".encode()
            )
            with open(self.index_path, "wb") as index_data_file:
                index_data_file.write(index_content)
        finally:
            self.index_write_lock.release()


class Debouncer:
    def __init__(self, debounce_period: float = 1.0, sleep_duration: float = 0.1):
        self.debounce_period = debounce_period
        self.sleep_duration = sleep_duration
        self.lock = threading.Lock()
        self.directories: list[Directory] = []
        self.flags = []
        self.running = True
        self.worker = threading.Thread(target=self.target)
        self.worker.daemon = True
        self.worker.start()

    def register(self, directory: Directory) -> int:
        with self.lock:
            self.directories.append(directory)
            self.flags.append(False)
            return len(self.directories) - 1

    def set(self, id: int):
        with self.lock:
            self.flags[id] = True

    def target(self):
        """Worker thread implementation."""
        next_save = time.monotonic()
        while True:
            now = time.monotonic()
            running = copy.copy(self.running)
            if now >= next_save or not running:
                directories_to_update: list[Directory] = []
                with self.lock:
                    for index, flag in enumerate(self.flags):
                        if flag:
                            directories_to_update.append(self.directories[index])
                            self.flags[index] = False
                for directory in directories_to_update:
                    directory.save_index_data(index_update_lock_already_acquired=False)
                next_save += self.debounce_period
                if not running:
                    break
            time.sleep(self.sleep_duration)

    def close(self):
        """Terminates the debouncer and perform all pending updates.

        This function is called automatically if Debouncer is used as a context manager.
        """
        self.running = False
        self.worker.join()

    def __enter__(self) -> "Debouncer":
        """Enables the use of the "with" statement.

        Returns:
            Debouncer: A debouncer context that calls :py:meth:`close` on exit.
        """
        return self

    def __exit__(
        self,
        type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        """Enables the use of the "with" statement.

        Args:
            type (typing.Optional[typing.Type[BaseException]]): None if the context exits without an exception, and the raised exception's class otherwise.
            value (typing.Optional[BaseException]): None if the context exits without an exception, and the raised exception otherwise.
            traceback (typing.Optional[types.TracebackType]): None if the context exits without an exception, and the raised exception's traceback otherwise.
        """
        self.close()


class BaseFile:
    """BaseFile implements the logic common to :py:class:`File` and :py:class:`OtherFile`.

    Args:
        directory (Directory): directory that contains the file.
    """

    def __init__(
        self, directory: Directory, name: str, metadata: dict[str, typing.Any]
    ):
        self._directory = directory
        self._path = self._directory.path / f"{name}.br"
        self._write_path = self._directory.path / f"{name}.br.write"
        self._handle = open(self._write_path, "wb")
        self._compressor = brotli.Compressor(
            mode=brotli.MODE_GENERIC, quality=11, lgwin=22, lgblock=0
        )
        self._index_entry = {
            "compressions": [
                {
                    "hash": None,
                    "size": 0,
                    "suffix": ".br",
                    "type": "brotli",
                }
            ],
            "hash": None,
            "metadata": metadata,
            "name": name,
            "size": 0,
        }
        self.uncompressed_hash_object = undr.utilities.new_hash()
        self.compressed_hash_object = undr.utilities.new_hash()

    def close(self):
        """Closes the file, moves it to its final location, and adds it to the index.

        This function is called automatically if BaseFile is used as a context manager.
        """
        if self._handle is None or self._compressor is None:
            raise Exception(f"close called twice for file {self._path}")
        compressed_data = self._compressor.finish()
        self._handle.write(compressed_data)
        self._handle.close()
        self._write_path.replace(self._path)
        self._index_entry["compressions"][0]["size"] += len(compressed_data)
        self.compressed_hash_object.update(compressed_data)
        self._index_entry["hash"] = self.uncompressed_hash_object.hexdigest()
        self._index_entry["compressions"][0][
            "hash"
        ] = self.compressed_hash_object.hexdigest()
        self._directory.update_index(self)
        self._handle = None
        self._compressor = None

    def write_raw(self, data: bytes):
        """Compresses and writes bytes to the file.

        This function may be called any number of times until the file is closed.

        Args:
            data (bytes): Uncompressed data.
        """
        assert self._handle is not None
        assert self._compressor is not None
        compressed_data = self._compressor.process(data)
        self._handle.write(compressed_data)
        self._index_entry["size"] += len(data)
        self._index_entry["compressions"][0]["size"] += len(compressed_data)
        self.uncompressed_hash_object.update(data)
        self.compressed_hash_object.update(compressed_data)


class File(BaseFile):
    def __init__(
        self,
        directory: Directory,
        type: Type,
        name: str,
        metadata: dict[str, typing.Any],
    ):
        super().__init__(
            directory=directory, name=f"{name}.{type.extension()}", metadata=metadata
        )
        self._index_entry["properties"] = type.properties()
        self.dtype = type.dtype()

    def write(self, data: numpy.ndarray):
        """Writes a numpy array to the file.

        This function may be called any number of times until the file is closed.

        Args:
            data (numpy.ndarray): structured numpy array whose dtype is identical to ``self.dtype``.
        """
        assert data.dtype == self.dtype
        self.write_raw(data.tobytes())

    def __enter__(self) -> "File":
        """Enables the use of the "with" statement.

        Returns:
            BaseFile: A file context that calls :py:meth:`close` on exit.
        """
        return self

    def __exit__(
        self,
        type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        """Enables the use of the "with" statement.

        Args:
            type (typing.Optional[typing.Type[BaseException]]): None if the context exits without an exception, and the raised exception's class otherwise.
            value (typing.Optional[BaseException]): None if the context exits without an exception, and the raised exception otherwise.
            traceback (typing.Optional[types.TracebackType]): None if the context exits without an exception, and the raised exception's traceback otherwise.
        """
        if type is None:
            self.close()


class OtherFile(BaseFile):
    def __init__(
        self, directory: Directory, name: str, metadata: dict[str, typing.Any]
    ) -> None:
        super().__init__(directory=directory, name=name, metadata=metadata)

    def write(self, data: typing.Union[str, bytes]):
        """Writes a string or bytes to a file.

        This function may be called any number of times until the file is closed.

        Args:
            data (typing.Union[str, bytes]): string or bytes (strings are encoded with utf-8).
        """
        if isinstance(data, str):
            self.write_raw(data.encode())
        else:
            self.write_raw(data)

    def __enter__(self) -> "OtherFile":
        """Enables the use of the "with" statement.

        Returns:
            BaseFile: A file context that calls :py:meth:`close` on exit.
        """
        return self

    def __exit__(
        self,
        type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        """Enables the use of the "with" statement.

        Args:
            type (typing.Optional[typing.Type[BaseException]]): None if the context exits without an exception, and the raised exception's class otherwise.
            value (typing.Optional[BaseException]): None if the context exits without an exception, and the raised exception otherwise.
            traceback (typing.Optional[types.TracebackType]): None if the context exits without an exception, and the raised exception's traceback otherwise.
        """
        if type is None:
            self.close()



import dataclasses
import pathlib
import typing

import undr
import numpy


class IncompleteHeader(Exception):
    def __init__(self):
        super().__init__("EOF reached while reading the header")


def read_aerdat_header(data: bytes) -> tuple[str, bytes]:
    index = 0
    while True:
        if index >= len(data):
            raise IncompleteHeader()
        if data[index] == 35:  # 35 is '#'
            line_end = data.find(b"\n", index + 1)
            if line_end == -1:
                break
            if not data[index:line_end].isascii():
                break
            index = line_end + 1
        else:
            break
    return (data[:index].decode(), data[index:])


@dataclasses.dataclass
class Aerdat1:
    header: typing.Optional[str]
    dvs: typing.Optional[numpy.ndarray]


def read_aerdat1(path: pathlib.Path) -> Aerdat1:
    aerdat1 = Aerdat1(
        header=None,
        dvs=None,
    )
    with open(path, "rb") as input:
        data = input.read()
    header, data = read_aerdat_header(data)
    if len(header) > 0:
        aerdat1.header = header
    data = data[: (len(data) // 6) * 6]
    if len(data) > 0:
        parsed_data = numpy.frombuffer(
            data, dtype=[("d0", "u1"), ("d1", "u1"), ("t", ">u4")]
        )
        aerdat1.dvs = numpy.zeros(len(parsed_data), dtype=undr.raw.DVS_DTYPE)
        aerdat1.dvs["t"] = parsed_data["t"]
        aerdat1.dvs["x"] = 127 - numpy.right_shift(parsed_data["d1"], 1)
        aerdat1.dvs["y"] = numpy.bitwise_and(parsed_data["d0"], 0x7F)
        aerdat1.dvs["p"] = numpy.bitwise_or(
            numpy.right_shift(numpy.bitwise_and(parsed_data["d0"], 0b10000000), 6),
            numpy.bitwise_and(parsed_data["d1"], 0b1),
        )
    return aerdat1


@dataclasses.dataclass
class Aerdat2:
    header: typing.Optional[str]
    dvs: typing.Optional[numpy.ndarray]
    aps: typing.Optional[numpy.ndarray]
    imu: typing.Optional[numpy.ndarray]
    aps_corrupted: bool
    imu_corrupted: bool

    @staticmethod
    def extract_x(parsed_data: numpy.ndarray) -> numpy.ndarray:
        return numpy.bitwise_or(
            numpy.left_shift(
                numpy.bitwise_and(parsed_data["d1"].astype("<u2"), 0b111111), 4
            ),
            numpy.right_shift(parsed_data["d2"].astype("<u2"), 4),
        )

    @staticmethod
    def extract_y(parsed_data: numpy.ndarray) -> numpy.ndarray:
        return numpy.bitwise_or(
            numpy.left_shift(
                numpy.bitwise_and(parsed_data["d0"].astype("<u2"), 0b1111111), 2
            ),
            numpy.right_shift(parsed_data["d1"].astype("<u2"), 6),
        )


def read_aerdat2(path: pathlib.Path) -> Aerdat2:
    aedat2 = Aerdat2(
        header=None,
        dvs=None,
        aps=None,
        imu=None,
        aps_corrupted=False,
        imu_corrupted=False,
    )
    with open(path, "rb") as input:
        data = input.read()
    header, data = read_aerdat_header(data)
    if len(header) > 0:
        aedat2.header = header
    data = data[: (len(data) // 8) * 8]
    parsed_data = numpy.frombuffer(
        data,
        dtype=[("d0", "u1"), ("d1", "u1"), ("d2", "u1"), ("d3", "u1"), ("t", ">u4")],
    )
    if numpy.count_nonzero(numpy.diff(parsed_data["t"].astype(">i8")) < 0) > 0:
        parsed_data = numpy.sort(parsed_data, order=["t"], kind="stable")
    dvs_mask = numpy.right_shift(parsed_data["d0"], 7) == 0
    aps_mask = numpy.logical_and(
        numpy.logical_not(dvs_mask),
        numpy.bitwise_and(parsed_data["d2"], 0b1100) != 0b1100,
    )
    imu_mask = numpy.logical_and(
        numpy.logical_not(dvs_mask), numpy.logical_not(aps_mask)
    )
    if numpy.count_nonzero(dvs_mask) > 0:
        dvs_data = parsed_data[dvs_mask]
        events = numpy.zeros(len(dvs_data), dtype=undr.raw.DVS_DTYPE)
        events["t"] = dvs_data["t"]
        events["x"] = Aerdat2.extract_x(dvs_data)
        events["y"] = Aerdat2.extract_y(dvs_data)
        events["p"] = numpy.bitwise_or(
            numpy.right_shift(
                numpy.bitwise_and(dvs_data["d2"].astype("<u2"), 0b1000), 3
            ),
            numpy.right_shift(
                numpy.bitwise_and(dvs_data["d2"].astype("<u2"), 0b100), 1
            ),
        )
        aedat2.dvs = events
    if numpy.count_nonzero(aps_mask) > 0:
        aps_data = parsed_data[aps_mask]
        aps_reset = numpy.bitwise_and(aps_data["d2"], 0b1100) == 0b0000
        aps_signal = numpy.bitwise_and(aps_data["d2"], 0b1100) == 0b0100
        aps_x = Aerdat2.extract_x(aps_data)
        aps_y = Aerdat2.extract_y(aps_data)
        aps_sample = numpy.bitwise_or(
            numpy.left_shift(numpy.bitwise_and(aps_data["d2"].astype("<u2"), 0b11), 8),
            aps_data["d3"].astype("<u2"),
        )
        t_frames = []
        current_frame = numpy.zeros((240, 180), dtype="<u2")
        first_reset = -1
        last_reset = -1
        first_signal = -1
        last_signal = -1
        for index in range(0, len(aps_data)):
            x = aps_x[index]
            y = aps_y[index]
            if x < 240 and y < 180:
                if aps_reset[index]:
                    if first_reset == -1:
                        first_reset = aps_data["t"][index]
                    if x == 0 and y == 0:
                        if (
                            first_reset != -1
                            and last_reset != -1
                            and first_signal != -1
                            and last_signal != -1
                        ):
                            t_frames.append(
                                (
                                    numpy.copy(current_frame),
                                    first_reset,
                                    last_reset,
                                    first_signal,
                                    last_signal,
                                )
                            )
                        current_frame.fill(0)
                        first_reset = aps_data["t"][index]
                        last_reset = -1
                        first_signal = -1
                        last_signal = -1
                    else:
                        last_reset = aps_data["t"][index]
                    current_frame[x, y] = aps_sample[index]
                elif aps_signal[index]:
                    if first_signal == -1:
                        first_signal = aps_data["t"][index]
                    last_signal = aps_data["t"][index]
                    luma = current_frame[x, y]
                    sample = aps_sample[index]
                    if sample > luma:
                        luma = 0
                    else:
                        luma -= sample
                    current_frame[x, y] = luma
        if (
            first_reset != -1
            and last_reset != -1
            and first_signal != -1
            and last_signal != -1
        ):
            t_frames.append(
                (current_frame, first_reset, last_reset, first_signal, last_signal)
            )
        if len(t_frames) == 0:
            aedat2.aps_corrupted = True
        else:
            aedat2.aps = numpy.zeros(len(t_frames), dtype=undr.raw.aps_dtype(240, 180))
            aedat2.aps["width"].fill(240)
            aedat2.aps["height"].fill(180)
            for index in range(0, len(t_frames)):
                aedat2.aps["t"][index] = t_frames[index][3]
                aedat2.aps["begin_t"][index] = t_frames[index][1]
                aedat2.aps["end_t"][index] = t_frames[index][3]
                aedat2.aps["exposure_begin_t"][index] = t_frames[index][2]
                aedat2.aps["exposure_end_t"][index] = t_frames[index][4]
                aedat2.aps["pixels"][index] = t_frames[index][0]
    if numpy.count_nonzero(imu_mask) > 0:
        imu_data = parsed_data[imu_mask]
        timestamps, indices, counts = numpy.unique(
            imu_data["t"], return_index=True, return_counts=True
        )
        unique_mask = counts == 7
        if numpy.count_nonzero(unique_mask) > 0:
            samples = numpy.bitwise_or(
                numpy.left_shift(
                    numpy.bitwise_and(imu_data["d0"].astype("<u2"), 0b1111), 12
                ),
                numpy.bitwise_or(
                    numpy.left_shift(imu_data["d1"].astype("<u2"), 4),
                    numpy.right_shift(imu_data["d2"].astype("<u2"), 4),
                ),
            )
            types = numpy.right_shift(numpy.bitwise_and(imu_data["d0"], 0b1110000), 4)
            boundary_indices = indices[unique_mask]
            valid = True
            for offset in range(0, 7):
                if not numpy.all(types[boundary_indices + offset] == offset):
                    aedat2.imu_corrupted = True
                    valid = False
                    break
            if valid:
                aedat2.imu = numpy.zeros(
                    numpy.count_nonzero(unique_mask), dtype=undr.raw.IMU_DTYPE
                )
                aedat2.imu["t"] = timestamps[unique_mask]
                aedat2.imu["accelerometer_x"] = samples[boundary_indices]
                aedat2.imu["accelerometer_y"] = samples[boundary_indices + 1]
                aedat2.imu["accelerometer_z"] = samples[boundary_indices + 2]
                aedat2.imu["temperature"] = samples[boundary_indices + 3]
                aedat2.imu["gyroscope_x"] = samples[boundary_indices + 4]
                aedat2.imu["gyroscope_y"] = samples[boundary_indices + 5]
                aedat2.imu["gyroscope_z"] = samples[boundary_indices + 6]
    return aedat2
